/* *************************** NOTICE *************************************
 * In accordance to the text of the Apache License 2.0 this text notifies
 * you that this file was not included in the original distribution. For
 * more information on the changes see the file README.patches.
 ************************** END of NOTICE ********************************/

/* Regular Expression Support for Alpine
 *
 * Written by Eduardo Chappa <chappa@washington.edu>
 *	      http://staff.washington.edu/chappa/alpine/
 *
 * Initial Version	: September 18, 2007
 * Last Modified	: October 03, 2007
 *
 * Send all comments about this module to the address listed 
 * above, where you can also find updates to this module.
 */

#include "../pith/headers.h"
#include "../pith/regex.h"

/* internal functions */
int perstr (REXPATTERN_S **, REXSTRINGT_S *);
int peranycharrep (REXPATTERN_S **, REXSTRINGT_S *);
int percharany (REXPATTERN_S **, REXSTRINGT_S *);
int false_fcn (REXPATTERN_S **, REXSTRINGT_S *);
int true_fcn (REXPATTERN_S **, REXSTRINGT_S *);
int pergroup (REXPATTERN_S **, REXSTRINGT_S *);
void fix_pattern(REXPATTERN_S **);
int fixedstr (REXPATTERN_S **, REXSTRINGT_S *);
int variablestr (REXPATTERN_S **, REXSTRINGT_S *, REXPATTERN_S *);
REXPATTERN_S *parse_into_types(char *);
void clear_rexpattern(REXPATTERN_S **);
void parse_string_into_fixed_pattern(REXPATTERN_S **, REXSTRINGT_S *, int);
void parse_string_into_variable_pattern(REXPATTERN_S **, REXSTRINGT_S *);

void
clear_rexpattern(REXPATTERN_S **pattern)
{
  REXPATTERN_S *patd, *pat2;

  for(patd = *pattern; patd; patd = pat2->ornext){
    memset((void *)&patd->prcss, 0, sizeof(REXPPATT_S));
    for(pat2 = patd; pat2->next ; pat2 = pat2->next)
	memset((void *)&pat2->next->prcss, 0, sizeof(REXPPATT_S));
  }
}

#define PATP(T) ((REXPATTERN_S *)(*T))

void free_rexpattern(void **target)
{
  if(PATP(target)->ornext)
    free_rexpattern((void **) &(PATP(target))->ornext);

  if(PATP(target)->next)
    free_rexpattern((void **) &PATP(target)->next);

  memset((void *)&PATP(target)->prcss, 0, sizeof(REXPPATT_S));

  if(PATP(target)->value)
    fs_give((void **)&PATP(target)->value);

  fs_give((void **)target);
}

/* This function ends the process of parsing the pattern. We do this here
 * so that basic parsing be more simple.
 */
void
fix_pattern(REXPATTERN_S **patp)
{
   REXPATTERN_S *pat;
   for(pat = *patp; pat; pat = pat->next){
	switch(pat->ptype){
		case REXString		: pat->fcn = perstr; pat->wtype = REXFixed;
					  pat->len = strlen(pat->value);  break;
		case REXGroup		: pat->fcn = pergroup; pat->wtype = REXFixed;
					  pat->len = 1;break;
		case REXGlobalOne	: pat->fcn = percharany; pat->wtype = REXFixed;
					  pat->len = 1; break;
		case REXGlobalAny	: pat->fcn = strlen(pat->value) == 1
					     ? (*pat->value == REXANYCHR
						? true_fcn
						: peranycharrep)
					     : pergroup;
					  pat->wtype = REXVariable; pat->len = 0; break;
		default		: break;
	}
   }

   for((pat = *patp)->atype = REXLeft; pat && pat->next; pat=pat->next){
	pat->next->minpos = pat->minpos + pat->len;
	pat->next->atype  = (pat->ptype == REXGlobalAny
				&& strlen(pat->value) == 1
				&& *pat->value == REXANYCHR)
			    || pat->atype == REXRight
				 ? REXRight : REXLeft;
   }
}

void
parse_string_into_fixed_pattern(REXPATTERN_S **pattern, REXSTRINGT_S *pstring, 
								int total)
{
  REXPATTERN_S *pat, *p2;
  int length;

  for(pat = *pattern; pat ; pat = pat->next){
     if(pat->wtype == REXFixed){
	pat->prcss.bpmin = pstring->value + pat->minpos;
	pat->prcss.bpmax = pstring->value + pstring->len - total;
	for (length = 0, p2 = pat; p2 && p2->wtype == REXFixed; p2 = p2->next)
	   length += p2->len;
	total -= length;
	for(; pat && pat->next && pat->next->wtype == REXFixed; pat = pat->next);
     }
  }
}

void
parse_string_into_variable_pattern(REXPATTERN_S **pattern, REXSTRINGT_S *pstring)
{
  REXPATTERN_S *pat, *p2, *last = NULL;
  char *bp = NULL;

  for(pat = *pattern; pat ; pat = pat->next){
     if(pat->wtype == REXVariable){
	pat->prcss.bpmin = bp == NULL ? pstring->value : bp;
	for (p2 = pat; p2 && p2->wtype == REXVariable; p2 = p2->next);
	pat->prcss.bpmax = p2 ? p2->prcss.bp : pstring->value + pstring->len;
	bp = p2 ? p2->prcss.ep : pstring->value + pstring->len;
     } 
     last = pat;
  }
}


/* This function does all pre-processing parsing */
REXPATTERN_S *
parse_pattern_into_types(char *pattern)
{
  REXPATTERN_S *pat, *p2;
  char c, *p, *q;

  if(!pattern || !*pattern)
   return NULL;

  for (p = pattern; *p && *p == REXSEPARE; p++);
  for(q = p; *q && (*q != REXSEPARE || *(q-1) != REXQUOTEC); q++);
  c = *q;
  *q = '\0';
  if((pat = parse_into_types(p)) != NULL){
     fix_pattern(&pat);
     if(pat && c != '\0'){
       for(p2 = pat; p2->next; p2 = p2->next);
       p2->ornext = parse_pattern_into_types(q+1);
     }
  }
  *q = c;
  return pat;
}

/* This function receives a pattern and parses it into types of patterns.
 * On error, returns NULL
 */
REXPATTERN_S *
parse_into_types(char *pattern)
{
  REXPATTERN_S *ppt = NULL;
  char *p = pattern, *q, *cp, *t, *r;
  int state = REXINITIAL_ST;
  int error, done, ingroup = 0, curstate;
  REXPatType pt;

  if(!p || !*p)
    return NULL;

  curstate = REXINITIAL_ST;
  error = done = 0;
  q = cp = fs_get((strlen(pattern)+1)*sizeof(char));
  *cp = '\0';
  while(!error && !done){
  switch(*p){
	case REXQUOTEC : if(ingroup)
			*q++ = *p++;
		      else{
			switch(*(p+1)){
				case REXQUOTEC: case REXGLOBAL: case REXANYCHR:
				case REXOGROUP: case REXCGROUP: case REXLEFTPA:
				case REXRIGHTP:
					state = REXSTRING_ST; *q++ = *(p+1);
					p += 2; break;
				default: error++; break;
			}
		      }
		      break;

	case REXGLOBAL : if (ingroup)
			  *q++ = *p++;
		      else
			  state = REXGLOBAL_ST; 
		      break;

	case REXANYCHR : *q++ = *p++;
		      if(!ingroup){
			switch(*p){
				case REXGLOBAL: state = REXMATCHA_ST; break;
				default	      : state = REXMATCHO_ST; break;
			}
		      }
		      break;

	case REXOGROUP : if(*(p+1) && (t = strchr(p+2,']')) != NULL){
			  ingroup = 1;
			  r = p;	/* record the position of p */
			  state = REXBGROUP_ST;
		      }
		      else
			  error++;
		      break;

	case REXCGROUP : if(ingroup){	/* if we are collecting a group */
			if(p != r + 1)
			  state = REXEGROUP_ST;
			else{
			  state = REXSTRING_ST; /* set the new state, maybe */
			  *q++ = *p++;
			}
		      }
		      else
			error++;
		      break;

	default :     state = REXSTRING_ST;
		      if(!*p)
			state = REXFINAL_ST;
		      *q++ = *p++; 
		      break;
    }
    if(curstate != state){
	switch(curstate){
	     case REXINITIAL_ST: switch(state){
				case REXSTRING_ST: pt = REXString; break;
				case REXMATCHA_ST: pt = REXGlobalAny; p++;
						if(*p == '\0')
						  state = REXFINAL_ST;
						done++; 
						break;
				case REXMATCHO_ST: pt = REXGlobalOne;
						if(*p == '\0')
						  state = REXFINAL_ST;
						done++; 
						break;
				case REXFINAL_ST :
				case REXBGROUP_ST: pt = REXGroup; p++; break;
				case REXGLOBAL_ST:
				case REXEGROUP_ST: error++; break;
				default : break;
			      }
			      break;
	     case REXMATCHA_ST: 
	     case REXMATCHO_ST: 
			     switch(state){
				case REXSTRING_ST: p--;
				case REXFINAL_ST : done++; break;
				default: break;
			     }
			     break;

	     case REXBGROUP_ST: break;

	     case REXSTRING_ST: switch(state){
				case REXMATCHA_ST:
				case REXMATCHO_ST: --q; p--; done++; break;
				case REXGLOBAL_ST: p++;
						if(q == cp + 1)
						  pt = REXGlobalAny;
						else
						  p -= 2;
						if(*p == '\0')
						  state = REXFINAL_ST;
						done++;
						break;
				case REXEGROUP_ST: p++;
						if (*p == REXGLOBAL){
						   p++;
						   pt = REXGlobalAny;
						}
						if(*p == '\0')
						  state = REXFINAL_ST;
						done++;
						break;
				case REXBGROUP_ST: 
				case REXFINAL_ST : done++; break;
				default: break;
			     }
			     break;
		default: switch(curstate){
			    case REXEGROUP_ST: error++; break;
			    default: break;
			 }
			 done++;
			 break;
	}
	curstate = state;
    }
  }
  if(!error){
     *q = '\0';
     ppt = fs_get(sizeof(REXPATTERN_S));
     memset((void *)ppt, 0, sizeof(REXPATTERN_S));
     ppt->ptype = pt;
     ppt->value = cpystr(cp);
     ppt->fcn   = true_fcn;
     ppt->next  = state == REXFINAL_ST ? NULL : parse_into_types(p);
  }
  if(ppt && ppt->next == NULL && state != REXFINAL_ST) /* error in parsing */
     free_pattern((void **)&ppt);
  if(cp)
    fs_give((void **)&cp);
  return ppt;
}

/* This function is designed to be the control function for Variable width
 * search and match. The reason why we pass "last" is because pat is not an
 * array of pointers, so searching in it is harder. We avoid searching in this
 * case by passing the answer from the calling function. 
 */
int
variablestr (REXPATTERN_S **pattern, REXSTRINGT_S *pstring, REXPATTERN_S *last)
{
   REXPATTERN_S *pat;
   char *string, *bp;
   int rv;

   (*pattern)->prcss.bpmin = pstring->bp = last == NULL ? pstring->value : last->prcss.ep;
   for(pat = *pattern; pat && pat->wtype == REXVariable; pat = pat->next);
   (*pattern)->prcss.bpmax = pstring->ep = pat ? pat->prcss.bp : pstring->value + pstring->len;

   pstring->bp = (*pattern)->prcss.bpmin;
   for(pat = *pattern; pat && pat->wtype == REXVariable; pat = pat->next){
      if((rv = (pat->fcn)(&pat, pstring)) != 0)
   	pat->prcss.bp = pstring->bp;
      for(string = pat->prcss.bp+1; rv == 1 && 
   				string <= pstring->ep; string++){
   	pstring->bp = string;
   	bp = pat->prcss.bp;
        rv = (pat->fcn)(&pat, pstring);
   	pat->prcss.bp = bp;
      }
      pat->prcss.rv = pat->prcss.bp != pat->prcss.ep;
      pstring->bp = string - 1;
   }
   return (*pattern)->prcss.rv;
}

int
true_fcn (REXPATTERN_S **pattern, REXSTRINGT_S *pstring)
{
 (*pattern)->prcss.bp = pstring->bp;  
 (*pattern)->prcss.ep = pstring->ep;
 return (*pattern)->prcss.rv = 1;
}

/* This function finds a string repeated, searching from the beginning of the 
 * string
 */
int
peranycharrep (REXPATTERN_S **pattern, REXSTRINGT_S *pstring)
{
   (*pattern)->prcss.rv = *pstring->bp == *(*pattern)->value;
   (*pattern)->prcss.bp = pstring->bp;
   (*pattern)->prcss.ep = pstring->bp + (*pattern)->prcss.rv;
   return (*pattern)->prcss.rv;
}

/* This function is the control function for searh of strings that match the
 * pattern when such pattern has Fixed width.
 */
int
fixedstr (REXPATTERN_S **pattern, REXSTRINGT_S *pstring)
{
  REXPATTERN_S *pat = *pattern, *p2;
  int rv;
  char *first, *last, *string;

  if(pat->prcss.bpmin == NULL) 
     return pat->prcss.rv = 1; /* been there, done that */

  first = pat->atype == REXLeft ? pat->prcss.bpmin : pat->prcss.bpmax;
  last  = pat->atype == REXLeft ? pat->prcss.bpmax : pat->prcss.bpmin;
  for(string = first, rv = 0; rv == 0 
	&& ((pat->atype == REXLeft && string <= last) 
	   || (pat->atype == REXRight && string >= last));
	string += pat->atype == REXLeft ? 1 : -1){
    pstring->bp = string;
    rv = (pat->fcn)(&pat, pstring);
    if(rv == 1){
       pstring->bp = pat->prcss.ep;
       for(p2 = pat; rv == 1 && p2->next && p2->next->wtype == REXFixed; 
								p2 = p2->next){
          rv = (p2->next->fcn)(&p2->next, pstring);
          pstring->bp = p2->next->prcss.ep;
       }
    }
  }
  return (*pattern)->prcss.rv = rv;
}

int
perstr (REXPATTERN_S **pattern, REXSTRINGT_S *pstring)
{
  (*pattern)->prcss.rv = strstr(pstring->bp, (*pattern)->value) == pstring->bp;
  (*pattern)->prcss.bp = pstring->bp;
  (*pattern)->prcss.ep = pstring->bp + (*pattern)->len;
  return (*pattern)->prcss.rv;
}

/* This checks if a group is in a given string */
int
pergroup (REXPATTERN_S **pattern, REXSTRINGT_S *pstring)
{
  (*pattern)->prcss.rv = strchr((*pattern)->value, *pstring->bp) != NULL;
  (*pattern)->prcss.bp = pstring->bp;
  (*pattern)->prcss.ep = pstring->bp + (*pattern)->prcss.rv;
  return (*pattern)->prcss.rv;
}

/* this function checks if there is any character and ties the string
 * on the REXRight.
 */
int
percharany (REXPATTERN_S **pattern, REXSTRINGT_S *pstring)
{
   (*pattern)->prcss.rv = pstring->bp && *pstring->bp;
   (*pattern)->prcss.bp = pstring->bp;
   (*pattern)->prcss.ep = pstring->bp + (*pattern)->prcss.rv;
   return (*pattern)->prcss.rv;
}

int
process_pattern (REXPATTERN_S **pattern, REXSTRINGT_S *pstring)
{
  REXPATTERN_S *pat, *last = NULL;
  int rv;

		/* first do all Fixed width strings */
  for(rv = 1, pat = *pattern;  rv && pat ; pat = pat->next)
 	if(pat->wtype == REXFixed)
	   rv = fixedstr (&pat, pstring);

  if(rv == 0)
    return 0;

		/* Now do all Variable width strings */
  for(pat = *pattern;  pat ; last = pat, pat = pat->next)
 	if(pat->wtype == REXVariable)
	   variablestr (&pat, pstring, last);
  return 1;
}

int
result_pattern(REXPATTERN_S *pattern, char **bp, char **ep)
{
  int rv, done, donebv, donebf;
  char *fp = *bp, *lp = *ep;
  REXPATTERN_S *pat, *last = NULL, *patb;

  /* first find the first pointer */
  done = donebv = donebf = 0;
  for(pat = pattern; done == 0 && pat ; pat = pat->next){
	switch(pat->wtype){
		case REXFixed	: if(last == NULL)
				    fp = pat->prcss.bp;
				  else if(last->prcss.ep != pat->prcss.bp){
				    fp = donebf ? NULL : pat->prcss.bp;
				    donebf = 1;
				  }
				  done++;
				  break; 
		case REXVariable: if(donebv == 0)
				    fp = pat->prcss.bp;
				  donebv = 1;
		default		: break;
	}
	last = pat;
  }

  rv = fp != NULL;

  /* now find the last pointer */
  if(rv > 0){
    patb = last;
    lp = patb->prcss.ep;
    for(pat = patb, done = 0; done == 0 && pat && pat->next; pat = pat->next){
	if(pat->prcss.ep == pat->next->prcss.bp)
	   lp = pat->prcss.ep;
	else done++;
    }
    if(!done)
      lp = pat->prcss.ep;
    else
      rv = 0;
  }
  clear_rexpattern(&pattern);
  *bp = fp; *ep = lp;
  return rv;
}

/* Call this function to execute the pattern on the string. Use bom and
 * eom to find the beginning and end of a match, if any. The return value
 * is 1, if there is a match, and 0 if there was no match.
 */
int
execute_pattern_on_string(REXPATTERN_S *pat, char *string, int *bom, int *eom)
{
  REXPATTERN_S *patd, *pat2;
  int rv = 0, len;
  char *bp, *ep;
  REXSTRINGT_S  strparse;

  if(pat == NULL)
   return rv;

  strparse.value = strparse.bp = strparse.ep = string;
  strparse.len   = strlen(strparse.value);

  for(patd = pat; rv == 0 && patd; patd = pat2->ornext){
    for(pat2 = patd, len = pat2->len; pat2->next ; pat2 = pat2->next)
       len += pat2->next->len;

    if(len <= strparse.len){
      parse_string_into_fixed_pattern(&patd, &strparse, len);
      if((rv = process_pattern(&patd, &strparse)) != 0){
	parse_string_into_variable_pattern(&patd, &strparse);
	bp = ep = strparse.value;
	if((rv = result_pattern(patd, &bp, &ep)) != 0){
	  if(bom) *bom = bp - strparse.value;
	  if(eom) *eom = ep - strparse.value;
	}
      }
    }
  }
  return rv;
}
