/* *************************** NOTICE *************************************
 * In accordance to the text of the Apache License 2.0 this text notifies
 * you that this file was not included in the original distribution. For
 * more information on the changes see the file README.patches.
 ************************** END of NOTICE ********************************/

#ifndef REGEXTYPE_INCLUDED
#define REGEXTYPE_INCLUDED

typedef enum {REXString, REXGroup, REXGlobalAny, REXGlobalOne} REXPatType;
typedef enum {REXFixed, REXVariable, REXAtLeastOne, REXAtMostOne} REXWidthType;
typedef enum {REXLeft, REXRight} REXAlignType;


/* GlobalTAnyNumber means any number of characters must be matched.
   GlobalTOne matches any character, but only once
 */

typedef struct proc_patt {
   int  rv;			/* The return value of the process	*/
   char *bp;			/* The pointer to the begin match	*/
   char *ep;			/* The pointed to the end of match	*/
   char *bpmin;			/* the min position where string begins	*/
   char *bpmax;			/* the max position where string begins	*/
} REXPPATT_S;

typedef struct pattern_t {
   char *value;			/* pattern to match			*/
   REXPatType ptype;		/* type of pattern to process		*/
   REXWidthType wtype;		/* type of width, fixed, variable, etc.	*/
   REXAlignType atype;		/* look match from left or right	*/
   int (*fcn)();		/* function we will use to process	*/
   int len;			/* length of pattern to match		*/
   int minpos;			/* Minimum place to look match for	*/
   REXPPATT_S prcss;		/* result of the process to that pattern*/
   struct pattern_t *next;	/* next pattern	in same match		*/
   struct pattern_t *ornext;	/* or try this pattern to match		*/
} REXPATTERN_S;

typedef struct string_t {
   char *value;			/* The actual string			*/
   int  len;			/* The length of the string		*/
   char *bp;			/* the position where string begins	*/
   char *ep;			/* the position where string ends	*/
} REXSTRINGT_S;

#define REXINITIAL_ST	 0	/* We begin in this state		*/
#define REXSTRING_ST	 1	/* We are reading a literal string	*/
#define REXBGROUP_ST	 2	/* A group is coming up			*/
#define REXEGROUP_ST	 3	/* An end of group has been found	*/
#define REXMATCHA_ST	 4	/* match all state			*/
#define REXMATCHO_ST	 5	/* match one character state		*/
#define REXGLOBAL_ST	 6	/* The '*' character			*/
#define REXFINAL_ST	99	/* We are done with the parsing		*/

#define REXGLOBAL  '*'
#define REXOGROUP  '['
#define REXCGROUP  ']'
#define REXQUOTEC  '\\'
#define REXANYCHR  '.'
#define REXSEPARE  '|'
#define REXLEFTPA  '('
#define REXRIGHTP  ')'

#endif /* REGEXTYPE_INCLUDED */

